//: Playground - noun: a place where people can play


import UIKit

import PlaygroundSupport


// register custom font for playground.  (in app, use Info.plist)
let fontURL = Bundle.main.url(forResource: "SourceSansPro-Bold", withExtension: "otf")!
CTFontManagerRegisterFontsForURL(fontURL as CFURL, CTFontManagerScope.process, nil)

/**
 Draws a filled circle, a "satellite", located at the 3 o'clock position
 */
fileprivate class SatelliteView: UIView
{
  // API
  var satelliteRadius:CGFloat = 15 {
    didSet { self.setNeedsLayout() }
  }
  var color:UIColor = .black {
    didSet { self.shapeLayer.fillColor = color.cgColor  }
  }
  
  override static var layerClass:AnyClass {
    return CAShapeLayer.self
  }
  
  private var shapeLayer:CAShapeLayer {
    return self.layer as! CAShapeLayer
  }
  
  override func layoutSubviews() {
    super.layoutSubviews()
    self.shapeLayer.path = self.satellitePath()
  }
  /** Returns a CGPath for an x-offset circle representing the satellite, computed from the satellite layer's bounds.
   
   This path only needs to be recomputed if the view (and therefore the satellite layer) changes its size, not merely for angle changes.
   
   */
  private func satellitePath() -> CGPath {
    let center = CGPoint(x: self.layer.bounds.midX, y: self.layer.bounds.midY)
    let ovalWidth = self.layer.bounds.size.width - 2 * self.satelliteRadius
    let satelliteUnrotatedCenter = CGPoint(x: center.x + ovalWidth/2.0, y: center.y)
    var satelitteRect = CGRect(origin: satelliteUnrotatedCenter, size: .zero)
    satelitteRect = satelitteRect.insetBy(dx: -1 * satelliteRadius, dy: -1 * satelliteRadius)
    let satelittePath = UIBezierPath(ovalIn: satelitteRect).cgPath
    return satelittePath
  }
}

/**
 Draws a UIView with the following features:
 
 - a dashed oval defining an orbit
 - text in the center
 - an optional filled "satellite" on the circle, at a configurable angle

 User's can animate `satelliteAngle` using UIViewPropertyAnimator, since it is a
 facade to a plain old `UIView.transform` property
 
 ### Implementation
 
 Draws the dashed line with drawRect.
 
 Draws the satelitte with a dedicate UIView, and rotates it with a transform.
 
 */
final class CirclePacer : UIView
{
  // client API
  
  var shouldShowSatellite:Bool {
    get { return !self.satellite.isHidden }
    set { self.satellite.isHidden = !newValue }
  }
  
  /// angular position of orbiting "satellite" circle in radians, where 0 points right, and increase goes clockwise.
  var satellitteAngle:CGFloat = 0
    { didSet { self.updateSatelliteTransform() } }
  
  /// text to display in center of the circle
  var text:String {
    get { return self.label.text ?? "" }
    set {
      self.label.text = newValue
      self.setNeedsLayout()
    }
  }
  
  // compile-time configurations & armature
  
  private let strokeColor:UIColor = UIColor.blue
  private let satelliteRadius:CGFloat = 15
  private let label:UILabel = UILabel(frame: CGRect.zero)
  fileprivate let satellite = SatelliteView(frame:CGRect.zero)
  // UIKit lifecycle methods
  
  override init(frame: CGRect)
  {
    super.init(frame: frame)
    self.setup()
  }
  
  required init?(coder aDecoder: NSCoder)
  {
    super.init(coder:aDecoder)
    self.setup()
  }

  private func setup()
  {
    // setup satellite subview
    self.satellite.satelliteRadius = self.satelliteRadius
    self.satellite.color = self.strokeColor
    self.satellite.frame  = self.bounds
    self.addSubview(self.satellite)
    
    // setup label subview
    if let font = UIFont(name: "SourceSansPro-Bold", size: 65) {
      label.font = font
    }
    else {
      print("WARNING: custom font SourceSansPro-Bold not registered")
    }
    label.text = "1"
    label.sizeToFit()
    label.textColor = self.strokeColor
    self.addSubview(label)
    
    self.setNeedsLayout()
  }

  override func draw(_ rect: CGRect)
  {
    //// General Declarations
    guard let context = UIGraphicsGetCurrentContext() else { return }
    
    //// draw orbit
    let ovalPath = UIBezierPath(ovalIn: self.bounds.insetBy(dx: satelliteRadius, dy: satelliteRadius))
    self.strokeColor.setStroke()
    ovalPath.lineWidth = 3
    context.saveGState()
    context.setLineDash(phase: 0, lengths: [4,6])
    context.setLineCap(CGLineCap.round)
    ovalPath.stroke()
    
    context.restoreGState()
  }
  
  override func layoutSubviews()
  {
    // center label
    label.center = CGPoint(x:self.bounds.midX,y:self.bounds.midY)

    // position satellite layer and configure its path, but do not rotate
    self.satellite.bounds = CGRect(origin:.zero,size:self.bounds.size)
    self.satellite.center = CGPoint(x: self.bounds.midX,
                                    y: self.bounds.midY)
    // rotate as needed
    self.updateSatelliteTransform()
  }
  
  /**
   Update the affine transform that places the satellite at the correct angle
   */
  private func updateSatelliteTransform()
  {
    let r = CGAffineTransform.identity.rotated(by: satellitteAngle)
    self.satellite.transform = r
  }
}

extension CirclePacer
{
  /**
   Animates to specified angles at corresponding relative times, over duration.
   
   A relative time is a `Double` in 0...1.0.
   */
  func animate(toAnglesAtRelativeTimes angleTimes:[(CGFloat,Double)],
               duration:TimeInterval)
  {
    UIView.animateKeyframes(withDuration: duration,
                            delay: 0,
                            options: .beginFromCurrentState,
                            animations:
      {
        var lastRelativePosition:Double = 0
        for (angle,relativeEnd) in angleTimes
        {
          let relativeDuration = relativeEnd - lastRelativePosition
          UIView.addKeyframe(withRelativeStartTime: lastRelativePosition,
                             relativeDuration: relativeDuration,
                             animations:
            {
              self.satellitteAngle = angle
          })
          lastRelativePosition = relativeEnd
        }
    },
                            completion: nil)
  }

  /**
   Animates a full circle clockwise.
   */
  func animateFullCircle(withDuration duration:TimeInterval) {
    let TAU:CGFloat = CGFloat(Double.pi * 2.0)
    let fullCircle:[(CGFloat,Double)] = [(v.satellitteAngle + TAU / 4.0,     1.0/4.0),
                                         (v.satellitteAngle + TAU / 2.0,     2.0/4.0),
                                         (v.satellitteAngle + TAU * 3.0/4.0, 3.0/4.0),
                                         (v.satellitteAngle + TAU,           1.0)]
    self.animate(toAnglesAtRelativeTimes: fullCircle,
                 duration: duration)
  }

  func animateTurn(duration:TimeInterval) {
    let anim = CABasicAnimation(keyPath: "transform.rotation.z")
    anim.fromValue = 0
    anim.toValue = Double.pi * 2.0
    anim.duration = duration
    anim.isCumulative = true
    anim.repeatCount = 0
    self.satellite.layer.add(anim, forKey: "rot")
  }
}

let v = CirclePacer(frame: CGRect(origin: .zero, size: CGSize(width: 200, height: 200)))
v.satellitteAngle = CGFloat(0)

//let v = SatelliteView(frame: CGRect(origin: .zero, size: CGSize(width: 200, height: 200)))

v.clipsToBounds = false
v.backgroundColor = UIColor.white
let container = UIView(frame: CGRect.init(origin: .zero, size: CGSize(width:350,height:350)))
container.addSubview(v)
v.center = container.center
v.satellitteAngle = CGFloat(-Double.pi / 2)
PlaygroundPage.current.liveView = container

//
//let TAU:CGFloat = CGFloat(M_PI * 2.0)
//v.satellitteAngle = CGFloat(TAU / 4.0) * -1.0
//let fullCircle:[(CGFloat,Double)] = [(v.satellitteAngle + TAU / 4.0,     1.0/4.0),
//                                     (v.satellitteAngle + TAU / 2.0,     2.0/4.0),
//                                     (v.satellitteAngle + TAU * 3.0/4.0, 3.0/4.0),
//                                     (v.satellitteAngle + TAU,           1.0)]
//
//v.animate(toAnglesAtRelativeTimes: fullCircle, duration: 10.0)

//UIViewPropertyAnimator(duration: 10, curve: UIViewAnimationCurve.linear) {
//  v.satellite.transform
//}.startAnimation()
v.animateTurn(duration: 10)


DispatchQueue.main.asyncAfter(deadline: DispatchTime.now() + .seconds(2)) {
  print(container.value(forKey: "recursiveDescription")!)
}
